package de.tu_dresden.lat.capi.implicateMatching;

import org.semanticweb.owlapi.apibinding.OWLManager;
import org.semanticweb.owlapi.model.OWLOntology;
import org.semanticweb.owlapi.model.OWLOntologyCreationException;

import java.io.File;
import java.io.IOException;
import java.util.*;
import java.util.stream.Collectors;
import java.util.Collection;
import java.util.Optional;
public class ParseAndMatch {

    private final static String NON_REDUNANT = "remove-redundant-axioms";
    private final static String SIMPLIFY = "simplify-conjunctions";
    private final static String SEMANTICALLY_MINIMAL = "semantically-ordered";

    private static void printHelpMessageAndExit(){
        System.out.println("Usage: ");
        System.out.println(ParseAndMatch.class.getCanonicalName()
                + " CLAUSE_FILE [ONTOLOGY] ["+NON_REDUNANT+"] ["+SIMPLIFY+"] ["+SEMANTICALLY_MINIMAL+"]");
        System.exit(1);
    }

    public static void main(String[] args) throws IOException, OWLOntologyCreationException {

        Set<String> KNOWN_ARGUMENTS = new HashSet();
        Collections.addAll(KNOWN_ARGUMENTS, NON_REDUNANT, SIMPLIFY, SEMANTICALLY_MINIMAL);


        if(args.length==0){
            printHelpMessageAndExit();
        }

        String filename = args[0];

        Set<String> additionalArguments = Arrays.stream(args)
                .skip(2)
                .collect(Collectors.toSet());

        if(!KNOWN_ARGUMENTS.containsAll(additionalArguments)){
            additionalArguments.removeAll(KNOWN_ARGUMENTS);
            System.out.println("Unknown argument(s): "+additionalArguments);
            printHelpMessageAndExit();
        }

        PrimeImplicateParser parser = new PrimeImplicateParser();
        parser.parse(new File(filename));
        PositiveClauseCollection posClauses = parser.getPosClauses();
        System.out.println("Relevant positive ground clauses:");
        System.out.println("-----");
        System.out.println(posClauses);
        System.out.println("number of positive ground clauses: "+posClauses.size());
        System.out.println();
        System.out.println();

        Collection<NegativeGroundClause> negClauses = parser.getNegClauses();
        System.out.println("Relevant negative ground clauses:");
        System.out.println("-----");
        for(NegativeGroundClause cl:negClauses) {
            System.out.println(cl);
        }
        System.out.println("number of negative ground clauses: "+negClauses.size());
        System.out.println();
        System.out.println();


        SolutionGenerator solutionGenerator = new SolutionGenerator(posClauses);

        Collection<Solution> solutions = solutionGenerator.generateSolutions(
                negClauses);

        if(!additionalArguments.isEmpty()) {
            System.out.println("Additional post-processing...");

            OWLOntology ontology = OWLManager.createOWLOntologyManager()
                    .loadOntologyFromOntologyDocument(new File(args[1]));

            OWLConverter converter = new OWLConverter(ontology.getOWLOntologyManager().getOWLDataFactory(),
                    new File(args[1]+".nameMap"));

            PostProcessing postProcessing = new PostProcessing(ontology,converter);

            System.out.println("Simplifying solutions....");
            solutions = solutions.stream()
                    .map(x -> {
                        if(additionalArguments.contains(SIMPLIFY)) {
                            System.out.println("Simplify conjunctions in "+x);
                            return postProcessing.simplifyAxioms(x);
                        }
                        else
                            return x;
                    })
                    .map(x -> {
                        if(additionalArguments.contains(NON_REDUNANT)) {
                            System.out.println("Remove redundancies in "+x);
                            return postProcessing.removeRedundantAxioms(x);
                        }
                        else
                            return x;
                    })
                    .collect(Collectors.toSet());

            if(additionalArguments.contains(SEMANTICALLY_MINIMAL)) {
                System.out.println("Sorting solutions...");
                solutions = postProcessing.sortBySemanticMinimality(solutions);
            }
        }

        System.out.println("Solutions: ");
        System.out.println("===========");
        System.out.println();
        for(Solution solution:solutions){
            System.out.println(solution);
            System.out.println("------------");
        }
        System.out.println("number of solutions: "+solutions.size());

        Optional<Integer> largest = solutions.stream().map(s -> s.size()).max(Integer::compare);
        if(largest.isPresent())
            System.out.println("size of largest solution: " +largest.get());

        Optional<Integer> largestAxiom = solutions.stream()
                .flatMap(s -> s.subsumptions().stream())
                .map(s -> s.size())
                .max(Integer::compare);
        if(largestAxiom.isPresent())
            System.out.println("size of largest axiom: " + largestAxiom.get());
    }
}
