package de.tu_dresden.lat.capi.implicateMatching;

import com.google.common.collect.Multimap;
import de.tu_dresden.lat.capi.tools.StrictPartialOrderDAG;
import org.semanticweb.HermiT.Reasoner;
import org.semanticweb.elk.owlapi.ElkReasonerFactory;
import org.semanticweb.owlapi.apibinding.OWLManager;
import org.semanticweb.owlapi.model.*;
import org.semanticweb.owlapi.reasoner.InferenceType;
import org.semanticweb.owlapi.reasoner.OWLReasoner;

import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.util.*;
import java.util.stream.Collectors;

public class PostProcessing {

    private final OWLReasoner reasoner;
    private final OWLDataFactory factory;
    private final OWLOntology ontology;
    private final OWLConverter converter;

    public PostProcessing(OWLOntology ontology, OWLConverter converter) {

        ontology.removeAxioms(ontology.getAxioms(AxiomType.DISJOINT_CLASSES));

        reasoner = new ElkReasonerFactory().createReasoner(ontology);
//        reasoner = new Reasoner.ReasonerFactory().createReasoner(ontology);
        factory = ontology.getOWLOntologyManager().getOWLDataFactory();
        this.ontology=ontology;
        this.converter=converter;
    }

    public Solution removeOntologyEntailedAxioms(Solution solution) {
        Solution filtered = new Solution();

        solution.subsumptions()
                .stream()
                .filter(x -> reasoner.isEntailed(converter.asOWLAxiom(x)))
                .forEach(filtered::addSubsumption);

        return solution;
    }


    public Solution removeRedundantAxioms(Solution solution) {

        Set<Subsumption> forSolution = new HashSet<>();

        Set<OWLSubClassOfAxiom> toRemove = new HashSet<>();

        solution.subsumptions()
                .stream()
                .forEach(subsumption -> {
                    OWLSubClassOfAxiom ax = converter.asOWLAxiom(subsumption);
                    if(!reasoner.isEntailed(ax)){
                        ontology.addAxiom(ax);
                        for(Iterator<Subsumption> it = forSolution.iterator(); it.hasNext();){
                            OWLSubClassOfAxiom ax2 = converter.asOWLAxiom(it.next());
                            ontology.remove(ax2);
                            reasoner.flush();
                            if(reasoner.isEntailed(ax2)){
                                System.out.println(subsumption+" makes "+ax2+" redundant");
                                it.remove();
                            } else
                                ontology.add(ax2);
                        }
                        forSolution.add(subsumption);
                        toRemove.add(ax);
                    } else {
                        System.out.println(subsumption+ " is redundant in "+forSolution);
                    }
                });

        ontology.removeAxioms(toRemove);

        return solution;
    }
    public Solution simplifyAxioms(Solution solution) {
        Solution result = new Solution();
        solution.subsumptions()
                .stream()
                .map(this::simplifyConjunctions)
                .forEach(result::addSubsumption);
        return result;
    }

    public Subsumption simplifyConjunctions(Subsumption subsumption) {
        return new Subsumption(simplifyConjunctions(subsumption.getLHS()),
                simplifyConjunctions(subsumption.getRHS()));
    }

    /**
     * Returns a subset of solutions so that no solution entails the other,
     * keeping those solutions that would be entailed by others.
     */
    public List<Solution> sortBySemanticMinimality(Collection<Solution> solutions) {
        StrictPartialOrderDAG<Solution> dag =
                new StrictPartialOrderDAG<Solution>(
                        StrictPartialOrderDAG.makeStrict(
                                (a,b) -> entails(b,a)
                        )
                );

        solutions.forEach(dag::addElement);

        System.out.println(dag);
        //System.exit(1);

        return dag.linearlizeTopBottom();
    }

    public boolean entails(Solution solution1, Solution solution2) {
        Set<OWLSubClassOfAxiom> toRemove = solution1.subsumptions()
                .stream()
                .map(converter::asOWLAxiom)
                .filter(x -> !ontology.containsAxiom(x))
                .collect(Collectors.toSet())
                ;

        ontology.addAxioms(toRemove);

        reasoner.flush();

        boolean result = solution2.subsumptions()
                .stream()
                .map(converter::asOWLAxiom)
                .allMatch(reasoner::isEntailed);

        ontology.removeAxioms(toRemove);

        return result;
    }

    private Collection<ClassName> simplifyConjunctions(Collection<ClassName> classes) {
        Set<ClassName> result = new HashSet<>();

        Set<OWLClass> conj = classes.stream()
                .map(converter::asOWLClass)
                .collect(Collectors.toSet());

        System.out.println("Axioms in ontology: "+ontology.getAxiomCount());

        reasoner.flush();
        reasoner.precomputeInferences(InferenceType.CLASS_HIERARCHY);

        classes.forEach(cl -> {
            OWLClass ocl = converter.asOWLClass(cl);

            conj.remove(ocl);

            //reasoner.precomputeInferences(InferenceType.CLASS_ASSERTIONS);
            //reasoner.flush();
            //System.out.println(factory.getOWLSubClassOfAxiom(
            //        conjunction(conj),
            //        ocl
            //));
            //System.out.println(reasoner.isEntailed(factory.getOWLSubClassOfAxiom(
            //        conjunction(conj),
            //        ocl
            //)));
            //System.out.println("sat: "+reasoner.isSatisfiable(
            //        factory.getOWLObjectIntersectionOf(conjunction(conj),
            //                factory.getOWLObjectComplementOf(ocl))
            //));
            //ontology.add(factory.getOWLClassAssertionAxiom(conjunction(conj), factory.getOWLNamedIndividual(IRI.create("a"))));
            //ontology.add(factory.getOWLClassAssertionAxiom(factory.getOWLObjectComplementOf(ocl), factory.getOWLNamedIndividual(IRI.create("a"))));
            //reasoner.flush();
            //System.out.println("Consistent: "+reasoner.isConsistent());
            //System.out.println(reasoner.getTypes(factory.getOWLNamedIndividual(IRI.create("a"))));

            if(!ocl.isOWLThing() && !reasoner.isEntailed(factory.getOWLSubClassOfAxiom(
                    conjunction(conj),
                    ocl
            ))) {
                result.add(cl);
                conj.add(ocl);
            }
        });

        return result;
    }

    private OWLClassExpression conjunction(Collection<OWLClass> conj) {
        conj.remove(factory.getOWLThing());
        if(conj.isEmpty())
            return factory.getOWLThing();
        else if(conj.size()==1)
            return conj.iterator().next();
        else
            return factory.getOWLObjectIntersectionOf(conj);
    }

}
