package de.tu_dresden.lat.capi.tools;

import java.util.*;
import java.util.stream.Collectors;

public class StrictPartialOrderDAG<A> {

    private StrictPartialOrder order;
    private Set<A> roots = new HashSet<>();
    private Map<A, Set<A>> dag = new HashMap<>();

    private Set<A> elements = new HashSet<>();

    public StrictPartialOrderDAG(StrictPartialOrder<A> order) {
        this.order=order;
    }

    public List<A> linearizeLeftToRight() {
        ArrayList<A> result = new ArrayList<>(elements.size());
        Set<A> processed = new HashSet<>();


        Deque<A> stack = new LinkedList<A>(roots);
        while(!stack.isEmpty()){
            A next = stack.pollFirst();
            if(!processed.contains(next)) {
                result.add(next);
                dag.getOrDefault(next, Collections.emptySet())
                        .forEach(stack::addFirst);
                processed.add(next);
            }
        }

        return result;
    }

    @Override
    public String toString() {
        String prefix = "";

        return toString(prefix, roots);
    }

    private String toString(String prefix, Collection<A> elements) {
        String result = "";
        for(A element: elements) {
            result += prefix+element.toString().replace("\n", ", ")+"\n";
            if(dag.containsKey(element)){
                result += toString("+ "+prefix, dag.get(element));
            }
        }
        return result;
    }

    public List<A> linearlizeTopBottom() {
        ArrayList<A> result = new ArrayList<>(elements.size());
        Set<A> processed = new HashSet<>();

        Queue<A> queue = new LinkedList<A>(roots);
        while(!queue.isEmpty()){
            A next = queue.poll();
            if(!processed.contains(next)) {
                result.add(next);
                dag.getOrDefault(next, Collections.emptySet())
                        .forEach(queue::add);
                processed.add(next);
            }
        }

        return result;
    }

    public void addElement(A a){
        if(elements.contains(a))
            return;

        elements.add(a);


        if(roots.isEmpty()) {
            roots.add(a);
        } else {
            Set<A> after = roots.stream().filter(x -> order.before(a,x )).collect(Collectors.toSet());
            if(!after.isEmpty()){
                roots.removeAll(after);
                roots.add(a);
                dag.put(a, after);
            } else {
                Set<A> before = roots.stream()
                        .filter(x -> order.before(x,a))
                        .collect(Collectors.toSet());
                if(before.isEmpty()){
                    roots.add(a);
                } else {
                    before.forEach(x -> addBelow(a, x));
                }
            }
        }
    }

    private void addBelow(A element, A below) {
        if(!dag.containsKey(below)){
            dag.put(below, new HashSet<>());
            dag.get(below).add(element);
        } else {
            Set<A> after = dag.get(below)
                    .stream()
                    .filter(x -> order.before(element,x))
                    .collect(Collectors.toSet());
            if(!after.isEmpty()) {
                dag.get(below).removeAll(after);
                dag.get(below).add(element);
                dag.put(element, after);
            } else {
                Set<A> next = dag.get(below)
                        .stream()
                        .filter(x -> order.before(x, element))
                        .collect(Collectors.toSet());
                if (next.isEmpty()) {
                    dag.get(below).add(element);
                } else {
                    next.forEach(x -> addBelow(element, x));
                }
            }
        }
    }

    public static <A> StrictPartialOrder<A> makeStrict(StrictPartialOrder<A> nonStrict) {
        return new StrictPartialOrder<A>() {
            @Override
            public boolean before(A a, A b) {
                boolean ab = nonStrict.before(a,b);
                boolean ba = nonStrict.before(b,a);
                if(ab && ba)
                    return a.hashCode()>b.hashCode();
                else
                    return ab;
            }
        };
    }

    /**
     * I assume this to be irreflexive, transitive and assymetric
     * @param <A>
     */
    public static interface StrictPartialOrder<A> {
        boolean before(A a, A b);
    }
}
